"use strict";

var _bluebird = _interopRequireDefault(require("bluebird"));

var _oracledb = _interopRequireDefault(require("oracledb"));

var _logger = _interopRequireDefault(require("./logger"));

var _lodash = _interopRequireDefault(require("lodash"));

var _sage_model = _interopRequireDefault(require("../build/sage_model"));

var _sage_schema = _interopRequireDefault(require("../build/sage_schema"));

var _sage_util = _interopRequireDefault(require("../build/sage_util"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// https://github.com/oracle/node-oracledb/blob/master/doc/api.md#-3214-stmtcachesize
// Statement caching can be disabled by setting the size to 0.
_oracledb.default.stmtCacheSize = 0;

const knex = require('knex')({
  client: 'oracle'
});

class Sage {
  constructor(options = {}) {
    this.Schema = _sage_schema.default;
    this._pool = null;
    this._connectOptions = null;
    this._connectURI = null;
    this.models = {}; // all the models that have currently been instantiated

    this.debug = options.debug;
    this.knex = knex;
    this.util = _sage_util.default;
    this.oracledb = _oracledb.default;
    this.logger = _logger.default;
  }

  log(o) {
    _logger.default.warn('This is deprecated.');

    _logger.default.debug(o);
  }

  getConnection(options = {}) {
    const self = this;
    return new _bluebird.default((resolve, reject) => {
      if (options.transaction) {
        const connection = options.transaction.connection;
        connection.isSageTransaction = true;
        return resolve(connection);
      }

      self._pool.getConnection((err, connection) => {
        if (err) {
          _logger.default.error('Out of connections!', err);

          return reject(err);
        }

        return resolve(connection);
      });
    });
  } // Promise wrap oracle connection.commit
  // Commits operations in the connection, then releases it


  commit(connection) {
    return new _bluebird.default((resolve, reject) => {
      connection.commit((err, result) => {
        if (err) {
          _logger.default.error('Could not commit', err); // Do not return yet. Release connection first.

        }

        sage.releaseConnection(connection).then(() => {
          if (err) {
            return reject(err);
          }

          return resolve();
        }).catch(reject);
      });
    });
  } // Used by statics and methods to figure out what to do with a connection
  // after the operation is performed. If this connection is part of a transaction
  // it will not close the connection.


  afterExecuteCommitable(connection) {
    if (connection.isSageTransaction) {
      return _bluebird.default.resolve();
    }

    return sage.commit(connection);
  } // Used by statics and methods to figure out what to do with a connection
  // after the operation is performed


  afterExecute(connection) {
    if (connection.isSageTransaction) {
      return _bluebird.default.resolve();
    }

    return sage.releaseConnection(connection);
  }
  /**
   Create a sage transaction to perform several operations before commit.
    You can create transactions either invoking as a Promise, or by passing down
   a function.
    It is suggested to always pass down a function, as in a function you are forced
   to apply a `commit()` or `rollback()` in order to resolve the promise.
    The Promise style is available in the event you need a slightly different syntax.
    Function Style:
    sage.transaction(function(t) {
    User.create({ transaction: t }).then(function() {
      t.commit();
    });
   }).then(function() {
    // transaction done!
   });
    Promise Style:
    sage.transaction().then(function(t) {
    User.create({ transaction: t }).then(function() {
      t.commit();
    });
   });
     */


  transaction(fn) {
    const self = this;

    if (fn) {
      return new _bluebird.default((resolve, reject) => {
        self.getConnection().then(connection => {
          const transaction = {
            connection,

            commit() {
              sage.commit(this.connection).then(resolve).catch(reject);
            },

            rollback(transaction) {
              sage.releaseConnection(this.connection).then(resolve).catch(reject);
            }

          };
          fn(transaction);
        });
      });
    }

    return self.getConnection().then(connection => {
      const transaction = {
        connection,

        commit() {
          return sage.commit(this.connection);
        },

        rollback(transaction) {
          return sage.releaseConnection(this.connection);
        }

      };
      return transaction;
    });
  }

  releaseConnection(connection) {
    return new _bluebird.default((resolve, reject) => {
      connection.release(err => {
        if (err) {
          _logger.default.error('Problem releasing connection', err);

          reject(err);
        } else {
          resolve();
        }
      });
    });
  }

  get connection() {
    _logger.default.warn('Sage connection is deprecated since pools');

    throw 'errr';
    return false;
    const self = this;
  }

  model(name, schema) {
    if (!schema) {
      const model = this.models[name];

      if (model) {
        return model.model;
      }

      return null;
    }

    return (0, _sage_model.default)(name, schema, this);
  } // disconnect() {
  //   let self = this;
  //   if(self._pool) {
  //     return new Promise(function(resolve, reject) {
  //       self._pool.release(function(err) {
  //         if(err) {
  //           console.error(err.message);
  //           reject(err);
  //         } else {
  //           self._pool = null;
  //           resolve(true);
  //         }
  //       })
  //     })
  //   } else {
  //     // No active connection
  //     return new Promise(function(resolve, reject) {
  //       resolve(true);
  //     })
  //   }
  // }


  execute(sql, bindParams = [], options = {}) {
    const self = this;
    return self._execute(sql, bindParams, options);
  }

  _execute(sql, bindParams, options) {
    const self = this;

    _logger.default.debug(sql);

    options = _lodash.default.extend({
      maxRows: 100,
      outFormat: self.oracledb.OBJECT
    }, options);
    let connection;
    let results;
    return self.getConnection({
      transaction: options.transaction
    }).then(c => connection = c).then(() => connection.execute(sql, bindParams, options)).then(r => {
      // Lowercase all the object keys
      // This was just done to make the data more 'JS' friendly since
      // the database column names are all uppercase.
      results = _lodash.default.map(r.rows, row => // Return the row with lowercased keys
      _lodash.default.transform(row, (result, val, key) => {
        result[key.toLowerCase()] = val;
      }));

      if (!options.transaction) {
        return sage.releaseConnection(connection);
      }
    }).then(() => results) // Return the results
    .catch(err => {
      _logger.default.warn(err);

      if (options.transaction) {
        throw err;
      }

      return self.releaseConnection(connection).then(() => {
        throw err;
      }).catch(err => {
        throw err;
      });
    });
  }

  connect(uri, connectOptions = {}) {
    const self = this;
    return self._connect(uri, connectOptions);
  }

  _connect(uri, connectOptions) {
    const self = this;

    if (self._pool) {
      return _bluebird.default.resolve();
    } // You passed in some optoins. We save them so that if you call connect() without connectOptions
    // it will use them again


    if (uri) {
      self._connectURI = uri;
    }

    if (_lodash.default.size(connectOptions) > 0) {
      self._connectOptions = connectOptions;
    } // Load saved values if they exist


    if (self._connectOptions) {
      connectOptions = self._connectOptions;
    }

    if (self._connectURI) {
      uri = self._connectURI;
    } // Make a new connection


    let auth = {
      user: 'system',
      password: 'oracle',
      connectString: uri || '127.0.0.1:1521/xe',
      poolMin: connectOptions.poolMin,
      poolMax: connectOptions.poolMax
    };
    auth = _lodash.default.defaults(connectOptions, auth);
    return new _bluebird.default((resolve, reject) => {
      _oracledb.default.createPool(auth, (err, pool) => {
        if (err) {
          _logger.default.error(err);

          return reject(err);
        }

        self._pool = pool;
        return resolve();
      });
    });
  }

}

let sage = new Sage();
module.exports = sage;